package ladasha.processes;


import java.io.*;
import javax.sound.midi.*;
import java.util.*;
import ladasha.statistics.*;
import ladasha.io.*;

public class Humanizer {
  // singleton
  private static Humanizer instance = null;
  
  private Humanizer(File midiFile, HashMap<String,File> profiles) {
    this.midiFile = midiFile;
    this.profiles = profiles;
  }
  
  public static Humanizer getInstance(File midiFile, HashMap<String,File> profiles) {
    if (instance == null) instance = new Humanizer(midiFile, profiles);
    return instance;
  }
  
  
  
  // fields
  private File midiFile = null;
  private HashMap<String,File> profiles = null;
  private Sequence sequenceOrig = null;
  private Sequence sequenceNew = null;
  private Sequencer seq1 = null;

  private static final int VELO_MASK = 0b1001;
  private static final int AMP_MASK = 0b1010;
  private static final int TIME_MASK = VELO_MASK; // same events

  // methods
  public boolean go() {
    requestSequence();
    modifySequence();
    return validateResults();
  }
  
  public void previewStop() {
    if(seq1 == null) return; // throw exception here in the future
    try {
      seq1.stop();
    } catch (Exception ex) { ex.printStackTrace(); }
  }
  
  public void previewStart() {
    try{
      seq1 = MidiSystem.getSequencer();
      if(seq1 == null) {
        return; // throw exception here in the future
      } else { 
        // get resources to get midi
        seq1.open();
      }
      // sequencer gets file
      seq1.setSequence(sequenceNew);
      Thread t = new Thread(new Runnable() {
      	public void run() { 
      	  seq1.start();  
        }
      });
      t.start();
    } catch(Exception ex){
      ex.printStackTrace();
    }
  }

  public void saveFile(String filename) {
    if (sequenceNew == null) return; // throw exception here in the future
    try {
      MidiSystem.write(sequenceNew, 1, new File(filename));
    } catch (Exception ex) {
      ex.printStackTrace();
    }
  }
  
  private void requestSequence() {
    MidiFileHandler mhandle = MidiFileHandler.getInstance();
    if (!mhandle.loadMidiFile(this.midiFile))
      // throw exception here in the future
      return;
    sequenceOrig = mhandle.getMidiSequence();
  }
  
  private void modifySequence() {
    sequenceNew = sequenceOrig; // copy the original sequence
    for (String s : profiles.keySet())
      modifySequence(s, profiles.get(s));
  }
  
  private void modifySequence(String selection, File profile) {
    String attr = selection.substring(selection.length()-1, selection.length()).toUpperCase();
    int trackNum = Integer.parseInt(selection.substring(0,selection.length()-1));
    
    // requires Java 7 (faster, but uglier)
    switch (attr) {
      case "A":
        modifyAmplitude(trackNum, profile);
        break;
      case "T":
        modifyTiming(trackNum, profile);
        break;
      case "V":
        modifyVelocity(trackNum, profile);
        break;
      default:
        // throw exception here in the future
        return;
    }
    
    // non-Java 7 version (slower, cleaner, portable)
    /*
    if (attr.equalsIgnoreCase("A"))
      modifyAmplitude(trackNum, profile);
    else if (attr.equalsIgnoreCase("T"))
      modifyTiming(trackNum,profile);
    else if (attr.equalsIgnoreCase("V"))
      modifyVelocity(trackNum, profile);
    else
      // throw exception here in the future
      return;
    */
  }
  
  // MIDI Messages of interest:
  // XXXX is channel #
  // KKKKKKK is note #
  // VVVVVVV is velocity
  // PPPPPPP is key pressure
  // 1001XXXX 0KKKKKKK 0VVVVVVV -> Note On (for Velocity)
  // 1010XXXX 0KKKKKKK 0PPPPPPP -> Polyphonic Key Pressure (for Amplitude)
  // for timing, manipulate MidiEvent.setTick() directly
  
  private void modifyVelocity(int tracknum, File profile) {    
    // load profile
    // check if discrete
      // yes: new discrete rng
    // check if gaussian
      // yes:  new gaussian rng
    // neither
      // exception/return
    
    // get track by number
    Track trackold = sequenceNew.getTracks()[tracknum];
    if (trackold == null) return; // out of bounds
    sequenceNew.deleteTrack(trackold);
    Track tracknew = sequenceNew.createTrack();
    
    // copy events into tracknew from trackold, unless a velocity command, in which case vary.
    int numevents = trackold.size();
    for (int i = 0; i < numevents; i++) {
      MidiEvent event = trackold.get(i);
      MidiMessage message = event.getMessage();
      int statusbyte = message.getStatus() >> 4;
      if ((statusbyte & VELO_MASK) != VELO_MASK) // not a velocity command
        tracknew.add(event);
      else { // a velocity command
        // do some processing
        // ...
        tracknew.add(event);
      }
    }
  }
  
  private void modifyAmplitude(int tracknum, File profile) {
    // load profile
    // check if discrete
      // yes: new discrete rng
    // check if gaussian
      // yes:  new gaussian rng
    // neither
      // exception/return
    
    // get track by number
    Track trackold = sequenceNew.getTracks()[tracknum];
    if (trackold == null) return; // out of bounds
    sequenceNew.deleteTrack(trackold);
    Track tracknew = sequenceNew.createTrack();
    
    // copy events into tracknew from trackold, unless an amplitude command, in which case vary.
    int numevents = trackold.size();
    for (int i = 0; i < numevents; i++) {
      MidiEvent event = trackold.get(i);
      MidiMessage message = event.getMessage();
      int statusbyte = message.getStatus() >> 4;
      if ((statusbyte & AMP_MASK) != AMP_MASK) // not an amplitude command
        tracknew.add(event);
      else { // an amplitude command
        // do some processing
        // ...
        tracknew.add(event);
      }
    }
  }
  
  private void modifyTiming(int tracknum, File profile) {
    // load profile
    // check if discrete
      // yes: new discrete rng
    // check if gaussian
      // yes:  new gaussian rng
    // neither
      // exception/return
    
    // get track by number
    Track trackold = sequenceNew.getTracks()[tracknum];
    if (trackold == null) return; // out of bounds
    sequenceNew.deleteTrack(trackold);
    Track tracknew = sequenceNew.createTrack();
    
    // copy events into tracknew from trackold, unless a timing command, in which case vary.
    int numevents = trackold.size();
    for (int i = 0; i < numevents; i++) {
      MidiEvent event = trackold.get(i);
      MidiMessage message = event.getMessage();
      int statusbyte = message.getStatus() >> 4;
      if ((statusbyte & TIME_MASK) != TIME_MASK) // not a timing command
        tracknew.add(event);
      else { // a timing command
        // do some processing
        // ...
        tracknew.add(event);
      }
    }
  }
  
  private boolean validateResults() {
    long threshold = 5000000; // +/- 5 seconds
    if (Math.abs(sequenceOrig.getMicrosecondLength() -
                 sequenceNew.getMicrosecondLength()) <= threshold)
      return true;
    return false;
  }
  
}